@@ -1080,7 +1080,11 @@ none_to_stream(gx_device_pdf * pdev)
             es->procs.process = templat->process;
             es->strm = s;
             (*templat->set_defaults) ((stream_state *) st);
-            (*templat->init) ((stream_state *) st);
+            code = (*templat->init) ((stream_state *) st);
+            if (code < 0) {
+                gs_free_object(pdev->pdf_memory, st, "none_to_stream");
+                return code;
+            }
             pdev->strm = s = es;
         }
     }
@@ -1589,10 +1593,15 @@ pdf_begin_aside(gx_device_pdf * pdev, pdf_resource_t ** plist,
                 pdf_resource_type_t type)
 {
     long id = pdf_begin_separate(pdev, type);
+    int code = 0;
 
     if (id < 0)
         return (int)id;
-    return pdf_alloc_aside(pdev, plist, pst, ppres, id);
+    code = pdf_alloc_aside(pdev, plist, pst, ppres, id);
+    if (code < 0)
+        (void)pdf_end_separate(pdev, type);
+
+    return code;
 }
 
 /* Begin a resource of a given type. */
@@ -1906,8 +1915,10 @@ pdf_page_id(gx_device_pdf * pdev, int page_num)
         pdev->num_pages = new_num_pages;
     }
     if ((Page = pdev->pages[page_num - 1].Page) == 0) {
-        pdev->pages[page_num - 1].Page = Page =
-            cos_dict_alloc(pdev, "pdf_page_id");
+        pdev->pages[page_num - 1].Page = Page = cos_dict_alloc(pdev, "pdf_page_id");
+        if (Page == NULL) {
+            return 0;
+        }
         Page->id = pdf_obj_forward_ref(pdev);
     }
     return Page->id;
