@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2012 Artifex Software, Inc.
+/* Copyright (C) 2001-2022 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  7 Mt. Lassen Drive - Suite A-134, San Rafael,
-   CA  94903, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
+   CA 94945, U.S.A., +1(415)492-9861, for further information.
 */
 
 
@@ -54,10 +54,11 @@ process_composite_text(gs_text_enum_t *pte, void *vbuf, uint bsize)
     gs_char chr, char_code = 0x0badf00d, space_char = GS_NO_CHAR;
     int buf_index = 0;
     bool return_width = (penum->text.operation & TEXT_RETURN_WIDTH);
+    gx_path *path = gs_text_enum_path(penum);
 
     str.data = buf;
     if (return_width) {
-        code = gx_path_current_point(penum->path, &penum->origin);
+        code = gx_path_current_point(path, &penum->origin);
         if (code < 0)
             return code;
     }
@@ -140,8 +141,12 @@ process_composite_text(gs_text_enum_t *pte, void *vbuf, uint bsize)
             gs_matrix_multiply(&prev_font->FontMatrix, psmat, &fmat);
             out.index = 0; /* Note : we don't reset out.xy_index here. */
             code = pdf_process_string_aux(&out, &str, NULL, &fmat, &text_state);
-            if (code < 0)
+            if (code < 0) {
+                if (code == gs_error_undefined && new_font && new_font->FontType == ft_encrypted2)
+                /* Caused by trying to make a CFF font resource for ps2write, which doesn't support CFF, abort now! */
+		  return_error(gs_error_invalidfont);
                 return code;
+            }
             curr.xy_index = out.xy_index; /* pdf_encode_process_string advanced it. */
             if (out.index < str.size) {
                 gs_glyph glyph;
@@ -236,7 +241,7 @@ static const char *const standard_cmap_names[] = {
     /* The following were added in PDF 1.2. */
 
     "GB-EUC-H", "GB-EUC-V",
-    "GBpc-EUC-H"
+    "GBpc-EUC-H",
 
     "B5pc-H", "B5pc-V",
     "ETen-B5-H", "ETen-B5-V",
@@ -282,7 +287,11 @@ attach_cmap_resource(gx_device_pdf *pdev, pdf_font_resource_t *pdfont,
         if (pcmap->CMapName.size == strlen(*pcmn) &&
             !memcmp(*pcmn, pcmap->CMapName.data, pcmap->CMapName.size))
             break;
-    if (*pcmn == 0) {
+
+    /* For PDF/A we need to write out all non-identity CMaps
+     * first force the identity check.
+     */
+    if (*pcmn == 0 || pdev->PDFA != 0) {
         /*
          * PScript5.dll Version 5.2 creates identity CMaps with
          * instandard name. Check this specially here
@@ -291,7 +300,10 @@ attach_cmap_resource(gx_device_pdf *pdev, pdf_font_resource_t *pdfont,
          */
         is_identity = gs_cmap_is_identity(pcmap, font_index_only);
     }
-    if (*pcmn == 0 && !is_identity) {		/* not standard */
+    /* If the CMap is non-standard, or we are producing PDF/A, and its not
+     * an Identity CMap, then we need to emit it.
+     */
+    if ((*pcmn == 0  || pdev->PDFA != 0) && !is_identity) {		/* not standard */
         pcmres = pdf_find_resource_by_gs_id(pdev, resourceCMap, pcmap->id + font_index_only);
         if (pcmres == 0) {
             /* Create and write the CMap object. */
@@ -329,31 +341,52 @@ attach_cmap_resource(gx_device_pdf *pdev, pdf_font_resource_t *pdfont,
         }
     }
     if (pcmres || is_identity) {
-        uint size = pcmap->CMapName.size;
-        byte *chars = gs_alloc_string(pdev->pdf_memory, size,
-                                      "pdf_font_resource_t(CMapName)");
+        if (pdfont->u.type0.CMapName_data == NULL || pcmap->CMapName.size != pdfont->u.type0.CMapName_size ||
+            memcmp(pdfont->u.type0.CMapName_data, pcmap->CMapName.data, pcmap->CMapName.size))
+        {
+            byte *chars = gs_alloc_bytes(pdev->pdf_memory->non_gc_memory, pcmap->CMapName.size,
+                                          "pdf_font_resource_t(CMapName)");
 
-        if (chars == 0)
-            return_error(gs_error_VMerror);
-        memcpy(chars, pcmap->CMapName.data, size);
+            if (chars == 0)
+                return_error(gs_error_VMerror);
+            memcpy(chars, pcmap->CMapName.data, pcmap->CMapName.size);
+            if (pdfont->u.type0.CMapName_data != NULL)
+                gs_free_object(pdev->pdf_memory->non_gc_memory, pdfont->u.type0.CMapName_data, "rewriting CMapName");
+            pdfont->u.type0.CMapName_data = chars;
+            pdfont->u.type0.CMapName_size = pcmap->CMapName.size;
+        }
         if (is_identity)
             strcpy(pdfont->u.type0.Encoding_name,
                     (pcmap->WMode ? "/Identity-V" : "/Identity-H"));
         else
-            gs_sprintf(pdfont->u.type0.Encoding_name, "%ld 0 R",
-                    pdf_resource_id(pcmres));
-        pdfont->u.type0.CMapName.data = chars;
-        pdfont->u.type0.CMapName.size = size;
+            gs_snprintf(pdfont->u.type0.Encoding_name, sizeof(pdfont->u.type0.Encoding_name),
+                        "%ld 0 R", pdf_resource_id(pcmres));
     } else {
+        uint size = 0;
+
         if (!*pcmn)
             /* Should not be possible, if *pcmn is NULL then either
              * is_identity is true or we create pcmres.
              */
             return_error(gs_error_invalidfont);
 
-        gs_sprintf(pdfont->u.type0.Encoding_name, "/%s", *pcmn);
-        pdfont->u.type0.CMapName.data = (const byte *)*pcmn;
-        pdfont->u.type0.CMapName.size = strlen(*pcmn);
+        size = strlen(*pcmn);
+        if (pdfont->u.type0.CMapName_data == NULL || size != pdfont->u.type0.CMapName_size ||
+            memcmp(pdfont->u.type0.CMapName_data, *pcmn, size) != 0)
+        {
+            byte *chars = gs_alloc_bytes(pdev->pdf_memory->non_gc_memory, size, "pdf_font_resource_t(CMapName)");
+            if (chars == 0)
+                return_error(gs_error_VMerror);
+
+            memcpy(chars, *pcmn, size);
+
+            if (pdfont->u.type0.CMapName_data != NULL)
+                gs_free_object(pdev->pdf_memory->non_gc_memory, pdfont->u.type0.CMapName_data, "rewriting CMapName");
+            pdfont->u.type0.CMapName_data = chars;
+            pdfont->u.type0.CMapName_size = size;
+        }
+
+        gs_snprintf(pdfont->u.type0.Encoding_name, sizeof(pdfont->u.type0.Encoding_name), "/%s", *pcmn);
         pdfont->u.type0.cmap_is_standard = true;
     }
     pdfont->u.type0.WMode = pcmap->WMode;
@@ -372,7 +405,7 @@ static int estimate_fontbbox(pdf_text_enum_t *pte, gs_font_base *font,
         return_error(gs_error_undefined);
     if (pfmat == 0)
         pfmat = &font->FontMatrix;
-    m = ctm_only(pte->pis);
+    m = ctm_only(pte->pgs);
     m.tx = fixed2float(pte->origin.x);
     m.ty = fixed2float(pte->origin.y);
     gs_matrix_multiply(pfmat, &m, &m);
@@ -421,8 +454,9 @@ scan_cmap_text(pdf_text_enum_t *pte, void *vbuf)
         gs_font *subfont = NULL;
         gs_point wxy;
         bool font_change = 0;
+        gx_path *path = gs_text_enum_path(pte);
 
-        code = gx_path_current_point(pte->path, &pte->origin);
+        code = gx_path_current_point(path, &pte->origin);
         if (code < 0)
             return code;
         do {
@@ -438,6 +472,8 @@ scan_cmap_text(pdf_text_enum_t *pte, void *vbuf)
             break_xy_index = scan.xy_index;
             code = font->procs.next_char_glyph(&scan, &chr, &glyph);
             if (code == 2) {		/* end of string */
+                if (subfont == NULL)
+                    subfont = scan.fstack.items[scan.fstack.depth].font;
                 done = true;
                 break;
             }
@@ -468,7 +504,9 @@ scan_cmap_text(pdf_text_enum_t *pte, void *vbuf)
                         return code;
                     break;
                 }
-                case ft_user_defined: {
+                case ft_user_defined:
+                case ft_PDF_user_defined:
+                {
                     gs_string str1;
 
                     str1.data = NULL;
@@ -499,10 +537,10 @@ scan_cmap_text(pdf_text_enum_t *pte, void *vbuf)
                 return code;
             if (break_index > start_index && pdev->charproc_just_accumulated)
                 break;
-            if (subfont->FontType == ft_user_defined &&
+            if ((subfont->FontType == ft_user_defined || subfont->FontType == ft_PDF_user_defined )&&
                 (break_index > start_index || !pdev->charproc_just_accumulated) &&
                 !(pdsubf->u.simple.s.type3.cached[cid >> 3] & (0x80 >> (cid & 7)))) {
-                if (subfont0 && subfont0->FontType != ft_user_defined)
+                if (subfont0 && subfont0->FontType != ft_user_defined && subfont0->FontType != ft_PDF_user_defined)
                     /* This is hacky. By pretending to be in a type 3 font doing a charpath we force
                      * text handling to fall right back to bitmap glyphs. This is because we can't handle
                      * CIDFonts with mixed type 1/3 descendants. Ugly but it produces correct output for
@@ -510,7 +548,7 @@ scan_cmap_text(pdf_text_enum_t *pte, void *vbuf)
                      */
                     pdev->type3charpath = 1;
                 pte->current_font = subfont;
-                return gs_error_undefined;
+                return_error(gs_error_undefined);
             }
             if (subfont->FontType == ft_encrypted || subfont->FontType == ft_encrypted2) {
                 font_change = (subfont != subfont0 && subfont0 != NULL);
@@ -540,7 +578,7 @@ scan_cmap_text(pdf_text_enum_t *pte, void *vbuf)
                 }
                 if (cid >= char_cache_size || cid >= width_cache_size)
                     return_error(gs_error_unregistered); /* Must not happen */
-                if (pdsubf->FontType == ft_user_defined  || pdsubf->FontType == ft_encrypted ||
+                if (pdsubf->FontType == ft_user_defined  || pdsubf->FontType == ft_PDF_user_defined  || pdsubf->FontType == ft_encrypted ||
                                 pdsubf->FontType == ft_encrypted2) {
                 } else {
                     pdf_font_resource_t *pdfont;
@@ -554,20 +592,32 @@ scan_cmap_text(pdf_text_enum_t *pte, void *vbuf)
                     if (code < 0)
                         return code;
                     if (pdf_is_CID_font(subfont)) {
-                        if (subfont->procs.decode_glyph((gs_font *)subfont, glyph, -1) != GS_NO_CHAR) {
+                        /* Some Pscript5 output has non-identity mappings between character code and CID
+                         * and the GlyphNames2Unicode dictionary uses character codes, not glyph names. So
+                         * if we detect ths condition we cheat and claim not to be a CIDFont, so that the
+                         * decode_glyph procedure can use the character code to look up the GlyphNames2Unicode
+                         * dictionary. See bugs #696021, #688768 and #687954 for examples of the various ways
+                         * this code can be exercised.
+                         */
+                        if (chr == glyph - GS_MIN_CID_GLYPH)
+                            code = subfont->procs.decode_glyph((gs_font *)subfont, glyph, -1, NULL, 0);
+                        else
+                            code = subfont->procs.decode_glyph((gs_font *)subfont, glyph, chr, NULL, 0);
+                        if (code != 0)
                             /* Since PScript5.dll creates GlyphNames2Unicode with character codes
                                instead CIDs, and with the WinCharSetFFFF-H2 CMap
                                character codes appears different than CIDs (Bug 687954),
                                pass the character code intead the CID. */
                             code = pdf_add_ToUnicode(pdev, subfont, pdfont,
                                 chr + GS_MIN_CID_GLYPH, chr, NULL);
-                        } else {
+                        else {
                             /* If we interpret a PDF document, ToUnicode
                                CMap may be attached to the Type 0 font. */
                             code = pdf_add_ToUnicode(pdev, pte->orig_font, pdfont,
                                 chr + GS_MIN_CID_GLYPH, chr, NULL);
                         }
-                    } else
+                    }
+                    else
                         code = pdf_add_ToUnicode(pdev, subfont, pdfont, glyph, cid, NULL);
                     if (code < 0)
                         return code;
@@ -589,6 +639,32 @@ scan_cmap_text(pdf_text_enum_t *pte, void *vbuf)
                                       (unsigned int)(glyph - GS_MIN_CID_GLYPH),
                                       buf);
                             pdsubf->used[cid >> 3] |= 0x80 >> (cid & 7);
+                            if (pdev->PDFA != 0) {
+                                switch (pdev->PDFACompatibilityPolicy) {
+                                    /* Default behaviour matches Adobe Acrobat, warn and continue,
+                                     * output file will not be PDF/A compliant
+                                     */
+                                    case 0:
+                                    case 1:
+                                    case 3:
+                                        emprintf(pdev->memory,
+                                             "All used glyphs mst be present in fonts for PDF/A, reverting to normal PDF output.\n");
+                                        pdev->AbortPDFAX = true;
+                                        pdev->PDFA = 0;
+                                        break;
+                                    case 2:
+                                        emprintf(pdev->memory,
+                                             "All used glyphs mst be present in fonts for PDF/A, aborting conversion.\n");
+                                        return_error(gs_error_invalidfont);
+                                        break;
+                                    default:
+                                        emprintf(pdev->memory,
+                                             "All used glyphs mst be present in fonts for PDF/A, unrecognised PDFACompatibilityLevel,\nreverting to normal PDF output\n");
+                                        pdev->AbortPDFAX = true;
+                                        pdev->PDFA = 0;
+                                        break;
+                                }
+                            }
                         }
                         cid = 0, code = 1;  /* undefined glyph. */
                         notdef_subst = true;
@@ -601,6 +677,29 @@ scan_cmap_text(pdf_text_enum_t *pte, void *vbuf)
                         }
                     } else if (code < 0)
                         return code;
+                    if (glyph == GS_MIN_CID_GLYPH && pdev->PDFA != 0) {
+                        switch (pdev->PDFACompatibilityPolicy) {
+                            case 0:
+                            case 1:
+                            case 3:
+                                emprintf(pdev->memory,
+                                     "A CIDFont uses CID 0, which is not legal for PDF/A, reverting to normal PDF output.\n");
+                                pdev->AbortPDFAX = true;
+                                pdev->PDFA = 0;
+                                break;
+                            case 2:
+                                emprintf(pdev->memory,
+                                     "A CIDFont uses CID 0, which is not legal for PDF/A, aborting conversion.\n");
+                                return_error(gs_error_invalidfont);
+                                break;
+                            default:
+                                emprintf(pdev->memory,
+                                     "A CIDFont uses CID 0, which is not legal for PDF/A, unrecognised PDFACompatibilityLevel,\nreverting to normal PDF output\n");
+                                pdev->AbortPDFAX = true;
+                                pdev->PDFA = 0;
+                                break;
+                        }
+                    }
                     if ((code == 0 /* just copied */ || pdsubf->Widths[cid] == 0) && !notdef_subst) {
                         pdf_glyph_widths_t widths;
 
@@ -635,10 +734,17 @@ scan_cmap_text(pdf_text_enum_t *pte, void *vbuf)
                         w0[cid] = widths.Width.w;
                     }
                     if (pdsubf->u.cidfont.CIDToGIDMap != 0) {
+                        uint gid = 0;
                         gs_font_cid2 *subfont2 = (gs_font_cid2 *)subfont;
 
-                        pdsubf->u.cidfont.CIDToGIDMap[cid] =
-                        subfont2->cidata.CIDMap_proc(subfont2, glyph);
+                        gid = subfont2->cidata.CIDMap_proc(subfont2, glyph);
+
+                        /* If this is a TrueType CIDFont, check the GSUB table to see if there's
+                         * a suitable substitute glyph.
+                         */
+                        if (subfont2->FontType == ft_CID_TrueType)
+                            gid = subfont2->data.substitute_glyph_index_vertical((gs_font_type42 *)subfont, gid, subfont2->WMode, glyph);
+                        pdsubf->u.cidfont.CIDToGIDMap[cid] = gid;
                     }
                 }
                 if (wmode)
@@ -689,17 +795,18 @@ scan_cmap_text(pdf_text_enum_t *pte, void *vbuf)
                 str.data = scan.text.data.bytes;
                 pdsubf = pdsubf0;
                 pte->text.operation = save_op;
+                pte->text.data.glyphs = save_data;
             }
-            pte->current_font = subfont0;
             if (!subfont0 || !pdsubf0)
                 /* This should be impossible */
                 return_error(gs_error_invalidfont);
 
+            pte->current_font = subfont0;
             code = gs_matrix_multiply(&subfont0->FontMatrix, &font->FontMatrix, &m3);
             /* We thought that it should be gs_matrix_multiply(&font->FontMatrix, &subfont0->FontMatrix, &m3); */
             if (code < 0)
                 return code;
-            if (pdsubf0->FontType == ft_user_defined  || pdsubf->FontType == ft_encrypted ||
+            if (pdsubf0->FontType == ft_user_defined  || pdsubf0->FontType == ft_PDF_user_defined  || pdsubf->FontType == ft_encrypted ||
                 pdsubf->FontType == ft_encrypted2)
                     pdfont = pdsubf0;
             else {
@@ -831,6 +938,7 @@ process_cmap_text(gs_text_enum_t *penum, void *vbuf, uint bsize)
     int code;
     pdf_text_enum_t *pte = (pdf_text_enum_t *)penum;
     byte *save;
+    uint start = pte->index;
 
     if (pte->text.operation &
         (TEXT_FROM_ANY - (TEXT_FROM_STRING | TEXT_FROM_BYTES))
@@ -857,6 +965,7 @@ process_cmap_text(gs_text_enum_t *penum, void *vbuf, uint bsize)
     code = scan_cmap_text(pte, vbuf);
     gs_free_string(pte->memory, (byte *)pte->text.data.bytes,  pte->text.size, "pdf_text_process");
     pte->text.data.bytes = save;
+    pte->bytes_decoded = pte->index - start;
 
     if (code == TEXT_PROCESS_CDEVPROC)
         pte->cdevproc_callout = true;
@@ -924,7 +1033,9 @@ process_cid_text(gs_text_enum_t *pte, void *vbuf, uint bsize)
     for (font = scaled_font; font->base != font; )
         font = font->base;
     /* Compute the scaling matrix. */
-    gs_matrix_invert(&font->FontMatrix, &scale_matrix);
+    code = gs_matrix_invert(&font->FontMatrix, &scale_matrix);
+    if (code < 0)
+        return code;
     gs_matrix_multiply(&scale_matrix, &scaled_font->FontMatrix, &scale_matrix);
 
     /* Find or create the CIDFont resource. */
