@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2021 Artifex Software, Inc.
+/* Copyright (C) 2001-2023 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -499,6 +499,12 @@ gdev_psdf_get_param(gx_device *dev, char *Param, void *list)
         return(psdf_write_string_param(plist, "sRGBProfile",
                                         &pdev->params.sRGBProfile));
     }
+    if (strcmp(Param, ".AlwaysOutline") == 0) {
+        return(psdf_get_embed_param(plist, ".AlwaysOutline", &pdev->params.AlwaysOutline));
+    }
+    if (strcmp(Param, ".NeverOutline") == 0) {
+        return(psdf_get_embed_param(plist, ".NeverOutline", &pdev->params.NeverOutline));
+    }
     if (strcmp(Param, ".AlwaysEmbed") == 0) {
         return(psdf_get_embed_param(plist, ".AlwaysEmbed", &pdev->params.AlwaysEmbed));
     }
@@ -598,6 +604,16 @@ gdev_psdf_get_params(gx_device * dev, gs_param_list * plist)
     if (code < 0)
         return code;
 
+    /* Font outlining parameters */
+
+    code = psdf_get_embed_param(plist, ".AlwaysOutline", &pdev->params.AlwaysOutline);
+    if (code < 0)
+        return code;
+
+    code = psdf_get_embed_param(plist, ".NeverOutline", &pdev->params.NeverOutline);
+    if (code < 0)
+        return code;
+
     /* Font embedding parameters */
 
     code = psdf_get_embed_param(plist, ".AlwaysEmbed", &pdev->params.AlwaysEmbed);
@@ -1117,6 +1133,8 @@ gdev_psdf_put_params(gx_device * dev, gs_param_list * plist)
         params.ColorImage.ACSDict = params.ColorImage.Dict = 0;
         params.GrayImage.ACSDict = params.GrayImage.Dict = 0;
         params.MonoImage.ACSDict = params.MonoImage.Dict = 0;
+        params.AlwaysOutline.data = params.NeverOutline.data = NULL;
+        params.AlwaysOutline.size = params.NeverOutline.size = 0;
         params.AlwaysEmbed.data = params.NeverEmbed.data = 0;
         params.AlwaysEmbed.size = params.AlwaysEmbed.persistent = params.NeverEmbed.size = params.NeverEmbed.persistent = 0;
         params.PSPageOptions.data = NULL;
@@ -1223,6 +1241,18 @@ gdev_psdf_put_params(gx_device * dev, gs_param_list * plist)
         goto exit;
     }
 
+    /* Font outlining parameters */
+
+    ecode = psdf_put_embed_param(plist, "~AlwaysOutline", ".AlwaysOutline",
+                                 &params.AlwaysOutline, mem, ecode);
+    ecode = psdf_put_embed_param(plist, "~NeverOutline", ".NeverOutline",
+                                 &params.NeverOutline, mem, ecode);
+
+    if (ecode < 0) {
+        code = ecode;
+        goto exit;
+    }
+
     /* Font embedding parameters */
 
     ecode = psdf_put_embed_param(plist, "~AlwaysEmbed", ".AlwaysEmbed",
@@ -1254,13 +1284,17 @@ gdev_psdf_put_params(gx_device * dev, gs_param_list * plist)
 exit:
     if (!(pdev->params.LockDistillerParams && params.LockDistillerParams)) {
         /* Only update the device paramters if there was no error */
-        /* If we have any copied param_string_arrays, start by freeing them */
-        if (pdev->params.PSPageOptions.size && params.PSPageOptions.size) {
-            int ix;
+        /* Do not permit changes to pdev->Params.PSPageOptions, it doesn't make any sense */
+        if (pdev->params.PSPageOptions.size != 0) {
+            if (params.PSPageOptions.size != 0 && params.PSPageOptions.data != pdev->params.PSPageOptions.data) {
+                int ix;
 
-            for (ix = 0; ix < pdev->params.PSPageOptions.size;ix++)
-                gs_free_object(mem->non_gc_memory, (byte *)pdev->params.PSPageOptions.data[ix].data, "freeing old string array copy");
-            gs_free_object(mem->non_gc_memory, (byte *)pdev->params.PSPageOptions.data, "freeing old string array");
+                for (ix = 0; ix < pdev->params.PSPageOptions.size;ix++)
+                    gs_free_object(mem->non_gc_memory, (byte *)params.PSPageOptions.data[ix].data, "freeing old string array copy");
+                gs_free_object(mem->non_gc_memory, (byte *)params.PSPageOptions.data, "freeing old string array");
+            }
+            params.PSPageOptions.data = pdev->params.PSPageOptions.data;
+            params.PSPageOptions.size = pdev->params.PSPageOptions.size;
         }
         pdev->params = params;
     } else {
