@@ -1,4 +1,4 @@
-/* Copyright (C) 2020-2022 Artifex Software, Inc.
+/* Copyright (C) 2020-2023 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -29,6 +29,8 @@
 
 #include "gscoord.h"         /* For gs_currentmatrix */
 
+static int pdfi_get_named_dest(pdf_context *ctx, pdf_obj *Named, pdf_obj **Dest);
+
 static int pdfi_pdfmark_setparam_obj(pdf_context *ctx, pdf_obj *obj, gs_param_string *entry)
 {
     int code = 0;
@@ -50,7 +52,7 @@ static int pdfi_pdfmark_setparam_pair(pdf_context *ctx, pdf_name *Key, pdf_obj *
     int code = 0;
 
     /* Handle the Key */
-    if (Key->type != PDF_NAME) {
+    if (pdfi_type_of(Key) != PDF_NAME) {
         code = gs_note_error(gs_error_typecheck);
         goto exit;
     }
@@ -391,19 +393,50 @@ static int pdfi_pdfmark_add_Page_View(pdf_context *ctx, pdf_dict *link_dict, pdf
     code = pdfi_array_get_no_store_R(ctx, dest_array, 0, (pdf_obj **)&page_dict);
     if (code < 0) goto exit;
 
-    if (page_dict->type != PDF_DICT) {
+    if(pdfi_type_of(page_dict) == PDF_INT) {
+        page_num = ((pdf_num *)page_dict)->value.i;
+    } else {
+        if (pdfi_type_of(page_dict) != PDF_DICT) {
+            if (pdfi_type_of(page_dict) != PDF_NULL) {
+                code = gs_note_error(gs_error_typecheck);
+                goto exit;
+            }
+            page_num = 0;
+        } else {
+            /* Find out which page number this is */
+            code = pdfi_page_get_number(ctx, page_dict, &page_num);
+            if (code < 0) goto exit;
+        }
+    }
+
+    page_num += ctx->Pdfmark_InitialPage;
+
+    if (pdfi_type_of(page_dict) != PDF_NULL)
+        /* Add /Page key to the link_dict
+         * Of course pdfwrite is numbering its pages starting at 1, because... of course :(
+         */
+        code = pdfi_dict_put_int(ctx, link_dict, "Page", page_num+1);
+    else
+        code = pdfi_dict_put_int(ctx, link_dict, "Page", 0);
+
+    if (code < 0)
+        goto exit;
+
+    /* Get and check the destination type, again without storing the deref in the array.
+     * In this case a memory leak  can only happen if the destination
+     * is an invalid type, but lets take no chances.
+     */
+    code = pdfi_array_get_no_store_R(ctx, dest_array, 1, &temp_obj);
+    if (code < 0) goto exit;
+
+    if (pdfi_type_of(temp_obj) != PDF_NAME) {
+        pdfi_countdown(temp_obj);
+        temp_obj = NULL;
         code = gs_note_error(gs_error_typecheck);
         goto exit;
     }
-
-    /* Find out which page number this is */
-    code = pdfi_page_get_number(ctx, page_dict, &page_num);
-    if (code < 0) goto exit;
-
-    /* Add /Page key to the link_dict
-     * Of course pdfwrite is numbering its pages starting at 1, because... of course :(
-     */
-    code = pdfi_dict_put_int(ctx, link_dict, "Page", page_num+1);
+    pdfi_countdown(temp_obj);
+    temp_obj = NULL;
 
     /* Build an array for /View, out of the remainder of the Dest entry */
     array_size = pdfi_array_size(dest_array) - 1;
@@ -457,12 +490,12 @@ static int pdfi_pdfmark_handle_dest_names(pdf_context *ctx, pdf_dict *link_dict,
         /* Note: in current implementation, PDF_STRING and PDF_NAME have all the same
          * fields, but just in case that changes I treat them separately here.
          */
-        if (name->type == PDF_STRING && dest->type == PDF_STRING) {
+        if (pdfi_type_of(name) == PDF_STRING && pdfi_type_of(dest) == PDF_STRING) {
             if (!pdfi_string_cmp((pdf_string *)name, (pdf_string *)dest)) {
                 found = true;
                 break;
             }
-        } else if (name->type == PDF_NAME && dest->type == PDF_NAME) {
+        } else if (pdfi_type_of(name) == PDF_NAME && pdfi_type_of(dest) == PDF_NAME) {
             if (!pdfi_name_cmp((pdf_name *)name, (pdf_name *)dest)) {
                 found = true;
                 break;
@@ -482,18 +515,19 @@ static int pdfi_pdfmark_handle_dest_names(pdf_context *ctx, pdf_dict *link_dict,
     code = pdfi_array_get(ctx, Names, i+1, (pdf_obj **)&D_dict);
     if (code < 0) goto exit;
 
-    if (D_dict->type == PDF_DICT) {
-        /* Dict is supposed to contain key "D" with Dest array */
-        code = pdfi_dict_knownget_type(ctx, D_dict, "D", PDF_ARRAY, (pdf_obj **)&dest_array);
-        if (code <= 0) goto exit;
-    } else {
-        if (D_dict->type == PDF_ARRAY) {
+    switch (pdfi_type_of(D_dict)) {
+        case PDF_DICT:
+            /* Dict is supposed to contain key "D" with Dest array */
+            code = pdfi_dict_knownget_type(ctx, D_dict, "D", PDF_ARRAY, (pdf_obj **)&dest_array);
+            if (code <= 0) goto exit;
+            break;
+        case PDF_ARRAY:
             dest_array = (pdf_array *)D_dict;
             D_dict = NULL;
-        } else {
+            break;
+        default:
             code = gs_note_error(gs_error_typecheck);
             goto exit;
-        }
     }
 
     /* Process the dest_array to replace with /Page /View */
@@ -512,7 +546,7 @@ static int pdfi_pdfmark_handle_dest_names(pdf_context *ctx, pdf_dict *link_dict,
  */
 int pdfi_pdfmark_modDest(pdf_context *ctx, pdf_dict *link_dict)
 {
-    int code = 0;
+    int code = 0, code1 = 0;
     pdf_dict *Dests = NULL;
     pdf_obj *Dest = NULL;
     bool delete_Dest = true;
@@ -529,14 +563,13 @@ int pdfi_pdfmark_modDest(pdf_context *ctx, pdf_dict *link_dict)
     code = pdfi_dict_knownget_type(ctx, ctx->Root, "Names", PDF_DICT, (pdf_obj **)&Names_dict);
     if (code < 0) goto exit;
 
-    switch (Dest->type) {
+    switch (pdfi_type_of(Dest)) {
     case PDF_ARRAY:
         code = pdfi_pdfmark_add_Page_View(ctx, link_dict, (pdf_array *)Dest);
         if (code < 0) goto exit;
         break;
     case PDF_NAME:
-    case PDF_STRING:
-        if (Dest->type == PDF_NAME && Dests != NULL) {
+        if (Dests != NULL) {
             /* Case where it's a name to look up in Contents(Root) /Dests */
             code = pdfi_dict_get_by_key(ctx, Dests, (const pdf_name *)Dest, (pdf_obj **)&dest_array);
             if (code == gs_error_undefined) {
@@ -545,28 +578,51 @@ int pdfi_pdfmark_modDest(pdf_context *ctx, pdf_dict *link_dict)
                 goto exit;
             }
             if (code < 0) goto exit;
-            if (dest_array->type != PDF_ARRAY) {
+            if (pdfi_type_of(dest_array) != PDF_ARRAY) {
                 code = gs_note_error(gs_error_typecheck);
                 goto exit;
             }
             code = pdfi_pdfmark_add_Page_View(ctx, link_dict, dest_array);
             if (code < 0) goto exit;
-        } else if (Names_dict != NULL) {
+            break;
+        }
+        /* fallthrough */
+    case PDF_STRING:
+        if (Names_dict != NULL) {
             /* Looking in Catalog(Root) for /Names<</Dests<</Names [name dict array]>>>> */
             code = pdfi_dict_knownget_type(ctx, Names_dict, "Dests", PDF_DICT, (pdf_obj **)&Dests);
             if (code < 0) goto exit;
             if (code == 0) {
                 /* TODO: Not found -- not sure if there is another case here or not */
+                code = gs_note_error(gs_error_undefined);
                 goto exit;
             }
 
             code = pdfi_dict_knownget_type(ctx, Dests, "Names", PDF_ARRAY, (pdf_obj **)&Names);
             if (code < 0) goto exit;
             if (code == 0) {
-                /* TODO: Not found -- not sure if there is another case here or not */
-                goto exit;
-            }
-            code = pdfi_pdfmark_handle_dest_names(ctx, link_dict, Dest, Names);
+                code = pdfi_get_named_dest(ctx, Dest, (pdf_obj **)&dest_array);
+                if (code < 0)
+                    goto exit;
+                if (pdfi_type_of(dest_array) == PDF_DICT) {
+                    pdf_obj *D = NULL;
+                    /* Dict is supposed to contain key "D" with Dest array */
+                    code = pdfi_dict_knownget_type(ctx, (pdf_dict *)dest_array, "D", PDF_ARRAY, &D);
+                    if (code <= 0) goto exit;
+
+                    pdfi_countdown(dest_array);
+                    dest_array = (pdf_array *)D;
+                }
+                if (pdfi_type_of(dest_array) == PDF_ARRAY) {
+                    code = pdfi_pdfmark_add_Page_View(ctx, link_dict, dest_array);
+                    if (code < 0)
+                        goto exit;
+                } else {
+                    pdfi_set_error(ctx, 0, NULL, E_PDF_BAD_NAMED_DEST, "pdfi_pdfmark_modDest", NULL);
+                }
+                break;
+            } else
+                code = pdfi_pdfmark_handle_dest_names(ctx, link_dict, Dest, Names);
             if (code < 0) goto exit;
         } else {
             /* TODO: Ignore it -- flag a warning? */
@@ -580,8 +636,9 @@ int pdfi_pdfmark_modDest(pdf_context *ctx, pdf_dict *link_dict)
  exit:
     if (delete_Dest) {
         /* Delete the Dest key */
-        code = pdfi_dict_delete(ctx, link_dict, "Dest");
-        if (code < 0) goto exit;
+        code1 = pdfi_dict_delete(ctx, link_dict, "Dest");
+        if (code1 < 0 && code >= 0)
+            code = code1;
     }
     pdfi_countdown(Dest);
     pdfi_countdown(Dests);
@@ -591,6 +648,244 @@ int pdfi_pdfmark_modDest(pdf_context *ctx, pdf_dict *link_dict)
     return code;
 }
 
+static int pdfi_check_limits(pdf_context *ctx, pdf_dict *node, char *str, int len)
+{
+    int code = 0, min, i, len2 = 0;
+    pdf_array *Limits = NULL;
+    pdf_string *Str = NULL;
+    char *str2 = NULL;
+
+    code = pdfi_dict_get_type(ctx, node, "Limits", PDF_ARRAY, (pdf_obj **)&Limits);
+    if (code < 0)
+        goto error;
+
+    if (pdfi_array_size(Limits) != 2) {
+        /* Limits are not valid, just ignore them. The calling code will then check
+         * the Names array.
+         */
+        pdfi_set_warning(ctx, 0, NULL, W_PDF_BAD_TREE_LIMITS, "pdfi_check_limits", 0);
+        goto error;
+    }
+
+    code = pdfi_array_get_type(ctx, Limits, 0, PDF_STRING, (pdf_obj **)&Str);
+    if (code < 0)
+        goto error;
+
+    if (pdfi_type_of(Str) == PDF_NAME) {
+        code = pdfi_string_from_name(ctx, (pdf_name *)Str, &str2, &len2);
+        if (code < 0)
+            return code;
+    } else {
+        len2 = ((pdf_string *)Str)->length;
+        str2 = (char *)gs_alloc_bytes(ctx->memory, len2 + 1, "pdfi_check_limits");
+        if (str2 == NULL) {
+            code = gs_note_error(gs_error_VMerror);
+            goto error;
+        }
+         memcpy(str2, ((pdf_string *)Str)->data, len2);
+         str2[len2] = 0;
+    }
+
+    pdfi_countdown(Str);
+    Str = NULL;
+
+    min = len;
+    if (len2 < min)
+        min = len2;
+
+    for (i=0;i< min;i++) {
+        if (str[i] < str2[i]) {
+            code = gs_note_error(gs_error_undefined);
+            goto error;
+        }
+        if (str[i] != str2[i])
+            break;
+    }
+    if (i > min && len2 < Str->length) {
+        code = gs_note_error(gs_error_undefined);
+        goto error;
+    }
+    gs_free_object(ctx->memory, str2, "pdfi_check_limits");
+    str2 = NULL;
+
+    code = pdfi_array_get_type(ctx, Limits, 1, PDF_STRING, (pdf_obj **)&Str);
+    if (code < 0)
+        goto error;
+
+    if (pdfi_type_of(Str) == PDF_NAME) {
+        code = pdfi_string_from_name(ctx, (pdf_name *)Str, &str2, &len2);
+        if (code < 0)
+            return code;
+    } else {
+        len2 = ((pdf_string *)Str)->length;
+        str2 = (char *)gs_alloc_bytes(ctx->memory, len2 + 1, "pdfi_check_limits");
+        if (str2 == NULL) {
+            code = gs_note_error(gs_error_VMerror);
+            goto error;
+        }
+         memcpy(str2, ((pdf_string *)Str)->data, len2);
+         str2[len2] = 0;
+    }
+
+    pdfi_countdown(Str);
+    Str = NULL;
+
+    min = len;
+    if (len2 < min)
+        min = len2;
+
+    for (i=0;i< min;i++) {
+        if (str[i] > str2[i]) {
+            code = gs_note_error(gs_error_undefined);
+            goto error;
+        }
+        if (str[i] != str2[i])
+            break;
+    }
+
+    if (i > min && len > len2)
+        code = gs_note_error(gs_error_undefined);
+
+error:
+    gs_free_object(ctx->memory, str2, "pdfi_check_limits");
+    pdfi_countdown(Str);
+    pdfi_countdown(Limits);
+    return code;
+}
+
+static int pdfi_get_name_from_node(pdf_context *ctx, pdf_dict *node, char *str, int len, pdf_obj **Name, bool is_root)
+{
+    int i = 0, code = 0;
+    pdf_string *StrKey = NULL;
+    pdf_array *NamesArray = NULL;
+    pdf_dict *Kid = NULL;
+    bool known;
+
+    code = pdfi_loop_detector_mark(ctx);
+    if (code < 0)
+        return code;
+
+    code = pdfi_dict_known(ctx, node, "Names", &known);
+    if (code < 0)
+        goto error;
+
+    if (known) {
+        code = pdfi_dict_known(ctx, node, "Limits", &known);
+        if (code < 0)
+            goto error;
+
+        if (!known) {
+            if (!is_root)
+                /* No Limits array (a required entry, except in the Root), so just assume that the
+                 * string is in this node and check all the Names anyway
+                 */
+                pdfi_set_warning(ctx, 0, NULL, W_PDF_NO_TREE_LIMITS, "pdfi_get_name_from_node", 0);
+        } else {
+            code = pdfi_check_limits(ctx, node, str, len);
+            if (code < 0)
+                goto error;
+        }
+
+        code = pdfi_dict_get_type(ctx, node, "Names", PDF_ARRAY, (pdf_obj **)&NamesArray);
+        if (code < 0)
+            goto error;
+
+        if (pdfi_array_size(NamesArray) & 1)
+            pdfi_set_warning(ctx, 0, NULL, W_PDF_NAMES_ARRAY_SIZE, "pdfi_get_name_from_node", 0);
+
+        for (i = 0;i < pdfi_array_size(NamesArray) / 2; i++) {
+            code = pdfi_array_get_type(ctx, NamesArray, i * 2, PDF_STRING, (pdf_obj **)&StrKey);
+            if (code < 0)
+                goto error;
+
+            if (StrKey->length == len && memcmp((const char *)StrKey->data, str, len) == 0) {
+                code = pdfi_array_get(ctx, NamesArray, (i * 2) + 1, (pdf_obj **)Name);
+                goto error;
+            }
+            pdfi_countdown(StrKey);
+            StrKey = NULL;
+        }
+        pdfi_countdown(NamesArray);
+        NamesArray = NULL;
+    }
+
+    /* Either no Names array (initial node) or not in array */
+    code = pdfi_dict_get_type(ctx, node, "Kids", PDF_ARRAY, (pdf_obj **)&NamesArray);
+    if (code < 0)
+        goto error;
+
+    for (i = 0;i < pdfi_array_size(NamesArray); i++) {
+        code = pdfi_array_get_type(ctx, NamesArray, i, PDF_DICT, (pdf_obj **)&Kid);
+        if (code < 0)
+            goto error;
+
+        code = pdfi_get_name_from_node(ctx, Kid, str, len, Name, false);
+        pdfi_countdown(Kid);
+        Kid = NULL;
+        if (code == 0)
+            break;
+
+        if (code < 0) {
+            if (code == gs_error_undefined)
+                continue;
+            goto error;
+        }
+    }
+
+error:
+    pdfi_countdown(Kid);
+    pdfi_countdown(StrKey);
+    pdfi_countdown(NamesArray);
+    (void)pdfi_loop_detector_cleartomark(ctx);
+    return code;
+}
+
+static int pdfi_get_named_dest(pdf_context *ctx, pdf_obj *Named, pdf_obj **Dest)
+{
+    int code = 0, len = 0;
+    pdf_dict *Names = NULL, *Dests = NULL;
+    char *str = NULL;
+
+    code = pdfi_loop_detector_mark(ctx);
+    if (code < 0)
+        return code;
+
+    code = pdfi_dict_get_type(ctx, ctx->Root, "Names", PDF_DICT, (pdf_obj **)&Names);
+    if (code < 0)
+        goto error;
+
+    code = pdfi_dict_get_type(ctx, Names, "Dests", PDF_DICT, (pdf_obj **)&Dests);
+    if (code < 0)
+        goto error;
+
+    if (pdfi_type_of(Named) == PDF_NAME) {
+        code = pdfi_string_from_name(ctx, (pdf_name *)Named, &str, &len);
+        if (code < 0)
+            return code;
+    } else {
+        len = ((pdf_string *)Named)->length;
+        str = (char *)gs_alloc_bytes(ctx->memory, len + 1, "pdfi_get_named_dest");
+        if (str == NULL) {
+            code = gs_note_error(gs_error_VMerror);
+            goto error;
+        }
+        memcpy(str, ((pdf_string *)Named)->data, len);
+        str[len] = 0;
+    }
+
+    code = pdfi_get_name_from_node(ctx, Dests, str, len, Dest, true);
+
+error:
+    if (pdfi_type_of(Named) == PDF_NAME)
+        (void)pdfi_free_string_from_name(ctx, str);
+    else
+        gs_free_object(ctx->memory, str, "pdfi_get_named_dest");
+    pdfi_countdown(Names);
+    pdfi_countdown(Dests);
+    pdfi_loop_detector_cleartomark(ctx);
+    return code;
+}
+
 /* Special handling for "A" in Link annotations and Outlines
  * Will delete A if handled and if A_key is provided.
  */
@@ -607,7 +902,7 @@ int pdfi_pdfmark_modA(pdf_context *ctx, pdf_dict *dict)
     code = pdfi_dict_get_no_store_R(ctx, dict, "A", (pdf_obj **)&A_dict);
     if (code < 0) goto exit;
 
-    if (A_dict->type != PDF_DICT) {
+    if (pdfi_type_of(A_dict) != PDF_DICT) {
         /* Invalid AP, just delete it because I dunno what to do...
          * TODO: Should flag a warning here
          */
@@ -631,19 +926,36 @@ int pdfi_pdfmark_modA(pdf_context *ctx, pdf_dict *dict)
     if (code <= 0) goto exit;
     /* We only handle GoTo for now */
     if (pdfi_name_is(S_name, "GoTo")) {
-        code = pdfi_dict_knownget_type(ctx, A_dict, "D", PDF_ARRAY, (pdf_obj **)&D_array);
-        if (code == 0) goto exit;
-        if (code < 0) {
-            if (code == gs_error_typecheck) {
-                /* TODO: Are there other cases to handle?
-                 * Sample tests_private/pdf/sumatra/recursive_action_destinations.pdf
-                 * has a recursive destination that has an indirect ref here.  We return a
-                 * typecheck and that causes us to omit the whole thing, but is that
-                 * really the best treatment?
-                 */
-            }
+        code = pdfi_dict_knownget(ctx, A_dict, "D", (pdf_obj **)&D_array);
+        if (code <= 0)
+            goto exit;
+        if (pdfi_type_of(D_array) == PDF_STRING || pdfi_type_of(D_array) == PDF_NAME)
+        {
+            pdf_obj *Dest = NULL;
+
+            code = pdfi_get_named_dest(ctx, (pdf_obj *)D_array, &Dest);
+            if (code < 0)
+                goto exit;
+            pdfi_countdown(D_array);
+            D_array = NULL;
+            if (pdfi_type_of(Dest) != PDF_ARRAY) {
+                if (pdfi_type_of(Dest) != PDF_DICT) {
+                    pdfi_countdown(Dest);
+                    code = gs_note_error(gs_error_typecheck);
+                    goto exit;
+                }
+                code = pdfi_dict_knownget(ctx, (pdf_dict *)Dest, "D", (pdf_obj **)&D_array);
+                pdfi_countdown(Dest);
+                if (code <= 0)
+                    goto exit;
+            } else
+                D_array = (pdf_array *)Dest;
+        }
+        if (pdfi_type_of(D_array) != PDF_ARRAY) {
+            code = gs_note_error(gs_error_typecheck);
             goto exit;
         }
+
         /* Process the D array to replace with /Page /View */
         code = pdfi_pdfmark_add_Page_View(ctx, dict, D_array);
         if (code < 0) goto exit;
@@ -1005,8 +1317,8 @@ void pdfi_pdfmark_write_boxes(pdf_context *ctx, pdf_dict *page_dict)
         "CropBox", "BleedBox", "TrimBox", "ArtBox"
     };
 
-    /* If the device doesn't support pdfmar, exit now */
-    if (!ctx->device_state.writepdfmarks)
+    /* If the device doesn't support pdfmark, or the device modifies the Page Size (making the other boxes unreliable) exit now */
+    if (!ctx->device_state.writepdfmarks || ctx->device_state.ModifiesPageSize)
         return;
 
     /* If we are using somethign other than the MediaBox, don't send these */
@@ -1043,7 +1355,7 @@ void pdfi_pdfmark_write_boxes(pdf_context *ctx, pdf_dict *page_dict)
             pdf_array *new_array = NULL;
 
             /* Box is present in page dicitonayr, check it's an array */
-            if (o->type != PDF_ARRAY) {
+            if (pdfi_type_of(o) != PDF_ARRAY) {
                 pdfi_countdown(o);
                 continue;
             }
