@@ -103,7 +103,7 @@ static int pdfi_trans_set_mask(pdf_context *ctx, pdfi_int_gstate *igs, int color
     double f;
     gs_matrix save_matrix, GroupMat, group_Matrix;
     gs_transparency_mask_subtype_t subtype = TRANSPARENCY_MASK_Luminosity;
-    pdf_bool *Processed = NULL;
+    bool Processed, ProcessedKnown = 0;
     bool save_OverrideICC = gs_currentoverrideicc(ctx->pgs);
 
 #if DEBUG_TRANSPARENCY
@@ -114,28 +114,26 @@ static int pdfi_trans_set_mask(pdf_context *ctx, pdfi_int_gstate *igs, int color
     /* Following the logic of the ps code, cram a /Processed key in the SMask dict to
      * track whether it's already been processed.
      */
-    code = pdfi_dict_knownget_type(ctx, SMask, "Processed", PDF_BOOL, (pdf_obj **)&Processed);
-    if (code > 0 && Processed->value) {
+    code = pdfi_dict_knownget_bool(ctx, SMask, "Processed", &Processed);
+    if (code > 0) {
+        if (Processed) {
 #if DEBUG_TRANSPARENCY
-        dbgmprintf(ctx->memory, "SMask already built, skipping\n");
+          dbgmprintf(ctx->memory, "SMask already built, skipping\n");
 #endif
-        goto exit;
+          code = 0;
+          goto exit;
+        }
+        ProcessedKnown = 1;
     }
 
     gs_setoverrideicc(ctx->pgs, true);
 
     /* If /Processed not in the dict, put it there */
     if (code == 0) {
-        /* the cleanup at end of this routine assumes Processed has a ref */
-        code = pdfi_object_alloc(ctx, PDF_BOOL, 0, (pdf_obj **)&Processed);
-        if (code < 0)
-            goto exit;
-        Processed->value = false;
-        /* pdfi_object_alloc() doesn't grab a ref */
-        pdfi_countup(Processed);
-        code = pdfi_dict_put(ctx, SMask, "Processed", (pdf_obj *)Processed);
+        code = pdfi_dict_put_bool(ctx, SMask, "Processed", false);
         if (code < 0)
             goto exit;
+        ProcessedKnown = 1;
     }
 
     /* See pdf1.7 pg 553 (pain in the butt to find this!) */
@@ -144,9 +142,12 @@ static int pdfi_trans_set_mask(pdf_context *ctx, pdfi_int_gstate *igs, int color
         /* G is transparency group XObject (required) */
         code = pdfi_dict_knownget_type(ctx, SMask, "G", PDF_STREAM, (pdf_obj **)&G_stream);
         if (code <= 0) {
-            dmprintf(ctx->memory, "WARNING: Missing 'G' in SMask, ignoring.\n");
             pdfi_trans_end_smask_notify(ctx);
-            code = 0;
+            if (ctx->args.pdfstoponerror)
+                code = gs_note_error(gs_error_undefined);
+            else
+                code = 0;
+            pdfi_set_error(ctx, 0, NULL, E_PDF_SMASK_MISSING_G, "pdfi_trans_set_mask", "");
             goto exit;
         }
 
@@ -157,36 +158,40 @@ static int pdfi_trans_set_mask(pdf_context *ctx, pdfi_int_gstate *igs, int color
         /* S is a subtype name (required) */
         code = pdfi_dict_knownget_type(ctx, SMask, "S", PDF_NAME, (pdf_obj **)&S);
         if (code <= 0) {
-            dmprintf(ctx->memory, "WARNING: Missing 'S' in SMask (defaulting to Luminosity)\n");
             subtype = TRANSPARENCY_MASK_Luminosity;
+            pdfi_set_warning(ctx, 0, NULL, W_PDF_SMASK_MISSING_S, "pdfi_trans_set_mask", "");
         }
         else if (pdfi_name_is(S, "Luminosity")) {
             subtype = TRANSPARENCY_MASK_Luminosity;
         } else if (pdfi_name_is(S, "Alpha")) {
             subtype = TRANSPARENCY_MASK_Alpha;
         } else {
-            dmprintf(ctx->memory, "WARNING: Unknown subtype 'S' in SMask (defaulting to Luminosity)\n");
             subtype = TRANSPARENCY_MASK_Luminosity;
+            pdfi_set_warning(ctx, 0, NULL, W_PDF_SMASK_UNKNOWN_S, "pdfi_trans_set_mask", "");
         }
 
         /* TR is transfer function (Optional) */
         code = pdfi_dict_knownget(ctx, SMask, "TR", (pdf_obj **)&TR);
         if (code > 0) {
-            if (TR->type == PDF_DICT || TR->type == PDF_STREAM) {
-                code = pdfi_build_function(ctx, &gsfunc, NULL, 1, TR, NULL);
-                if (code < 0)
-                    goto exit;
-                if (gsfunc->params.m != 1 || gsfunc->params.n != 1) {
-                    pdfi_free_function(ctx, gsfunc);
-                    gsfunc = NULL;
-                    dmprintf(ctx->memory, "WARNING: Ignoring invalid TR (number of inpuits or outputs not 1) in SMask\n");
-                }
-            } else if (TR->type == PDF_NAME) {
-                if (!pdfi_name_is((pdf_name *)TR, "Identity")) {
-                    dmprintf(ctx->memory, "WARNING: Unknown TR in SMask\n");
-                }
-            } else {
-                dmprintf(ctx->memory, "WARNING: Ignoring invalid TR in SMask\n");
+            switch (pdfi_type_of(TR)) {
+                case PDF_DICT:
+                case PDF_STREAM:
+                    code = pdfi_build_function(ctx, &gsfunc, NULL, 1, TR, NULL);
+                    if (code < 0)
+                        goto exit;
+                    if (gsfunc->params.m != 1 || gsfunc->params.n != 1) {
+                        pdfi_free_function(ctx, gsfunc);
+                        gsfunc = NULL;
+                        pdfi_set_warning(ctx, 0, NULL, W_PDF_SMASK_INVALID_TR, "pdfi_trans_set_mask", "");
+                    }
+                    break;
+                case PDF_NAME:
+                    if (!pdfi_name_is((pdf_name *)TR, "Identity")) {
+                        pdfi_set_warning(ctx, 0, NULL, W_PDF_SMASK_UNKNOWN_TR, "pdfi_trans_set_mask", "");
+                    }
+                    break;
+                default:
+                    pdfi_set_warning(ctx, 0, NULL, W_PDF_SMASK_UNKNOWN_TR_TYPE, "pdfi_trans_set_mask", "");
             }
         }
 
@@ -251,8 +256,14 @@ static int pdfi_trans_set_mask(pdf_context *ctx, pdfi_int_gstate *igs, int color
         if (code > 0) {
             /* TODO: Stuff with colorspace, see .execmaskgroup */
             code = pdfi_dict_knownget(ctx, Group, "CS", &CS);
-            if (code < 0)
-                goto exit;
+            if (code < 0) {
+                code = pdfi_dict_knownget(ctx, Group, "ColorSpace", &CS);
+                if (code < 0) {
+                    pdfi_set_error(ctx, 0, NULL, E_PDF_GROUP_NO_CS, "pdfi_trans_set_mask", (char *)"*** Defaulting to currrent colour space");
+                    goto exit;
+                }
+                pdfi_set_warning(ctx, 0, NULL, W_PDF_GROUP_HAS_COLORSPACE, "pdfi_trans_set_mask", NULL);
+            }
             if (code > 0) {
                 code = pdfi_create_colorspace(ctx, CS, (pdf_dict *)ctx->main_stream,
                                               ctx->page.CurrentPageDict, &pcs, false);
@@ -284,6 +295,9 @@ static int pdfi_trans_set_mask(pdf_context *ctx, pdfi_int_gstate *igs, int color
             }
             params.Background_components = pdfi_array_size(BC);
 
+            if (gs_color_space_num_components(params.ColorSpace) != params.Background_components)
+                pdfi_set_warning(ctx, 0, NULL, W_PDF_GROUP_BAD_BC, "pdfi_trans_set_mask", NULL);
+
             /* TODO: Not sure how to handle this...  recheck PS code (pdf_draw.ps/gssmask) */
             /* This should be "currentgray" for the color that we put in params.ColorSpace,
              * It looks super-convoluted to actually get this value.  Really?
@@ -298,9 +312,9 @@ static int pdfi_trans_set_mask(pdf_context *ctx, pdfi_int_gstate *igs, int color
             goto exit;
 
         code = pdfi_form_execgroup(ctx, ctx->page.CurrentPageDict, G_stream,
-                                   igs->GroupGState, NULL, &group_Matrix);
+                                   igs->GroupGState, NULL, NULL, &group_Matrix);
         code1 = gs_end_transparency_mask(ctx->pgs, colorindex);
-        if (code != 0)
+        if (code == 0)
             code = code1;
 
         /* Put back the matrix (we couldn't just rely on gsave/grestore for whatever reason,
@@ -309,11 +323,15 @@ static int pdfi_trans_set_mask(pdf_context *ctx, pdfi_int_gstate *igs, int color
         gs_setmatrix(ctx->pgs, &save_matrix);
 
         /* Set Processed flag */
-        if (code == 0 && Processed)
-            Processed->value = true;
+        if (code == 0 && ProcessedKnown)
+        {
+            code = pdfi_dict_put_bool(ctx, SMask, "Processed", true);
+            if (code < 0)
+                goto exit;
+        }
     } else {
         /* take action on a non-/Mask entry. What does this mean ? What do we need to do */
-        dmprintf(ctx->memory, "Warning: Type is not /Mask, entry ignored in pdfi_set_trans_mask\n");
+        pdfi_set_warning(ctx, 0, NULL, W_PDF_SMASK_UNKNOWN_TYPE, "pdfi_trans_set_mask", "");
     }
 
  exit:
@@ -332,7 +350,6 @@ static int pdfi_trans_set_mask(pdf_context *ctx, pdfi_int_gstate *igs, int color
     pdfi_countdown(BBox);
     pdfi_countdown(Matrix);
     pdfi_countdown(CS);
-    pdfi_countdown(Processed);
 #if DEBUG_TRANSPARENCY
     dbgmprintf(ctx->memory, "pdfi_trans_set_mask (.execmaskgroup) END\n");
 #endif
@@ -393,7 +410,7 @@ static int pdfi_transparency_group_common(pdf_context *ctx, pdf_dict *page_dict,
         /* Didn't find a /CS key, try again using /ColorSpace */
         code = pdfi_dict_knownget(ctx, group_dict, "ColorSpace", &CS);
     }
-    if (code > 0 && CS->type != PDF_NULL) {
+    if (code > 0 && pdfi_type_of(CS) != PDF_NULL) {
         code = pdfi_setcolorspace(ctx, CS, group_dict, page_dict);
         if (code < 0)
             goto exit;
@@ -703,7 +720,7 @@ int pdfi_trans_setup(pdf_context *ctx, pdfi_trans_state_t *state, gs_rect *bbox,
     bool current_overprint;
     bool okOPcs = false;
     bool ChangeBM = false;
-    gs_blend_mode_t mode;
+    gs_blend_mode_t  mode = gs_currentblendmode(ctx->pgs);  /* quite warning */
     bool need_group = false;
 
     memset(state, 0, sizeof(*state));
@@ -768,7 +785,12 @@ int pdfi_trans_setup(pdf_context *ctx, pdfi_trans_state_t *state, gs_rect *bbox,
         code = gs_setfillconstantalpha(ctx->pgs, 1.0);
         code = gs_setstrokeconstantalpha(ctx->pgs, 1.0);
     }
-    if (ChangeBM) {
+
+    /* If we are in a fill stroke situation, do not change the blend mode.
+       We can have situations where the fill and the stroke have different
+       blending needs (one may need compatible overprint and the other may
+       need the normal mode) This is all handled in the fill stroke logic */
+    if (ChangeBM && caller != TRANSPARENCY_Caller_FillStroke) {
         state->saveBM = mode;
         state->ChangeBM = true;
         code = gs_setblendmode(ctx->pgs, BLEND_MODE_CompatibleOverprint);
@@ -865,6 +887,7 @@ int pdfi_trans_teardown(pdf_context *ctx, pdfi_trans_state_t *state)
 
 int pdfi_trans_set_params(pdf_context *ctx)
 {
+    int code = 0;
     pdfi_int_gstate *igs = (pdfi_int_gstate *)ctx->pgs->client_data;
     gs_transparency_channel_selector_t csel;
 
@@ -874,9 +897,9 @@ int pdfi_trans_set_params(pdf_context *ctx)
         else
             csel = TRANSPARENCY_CHANNEL_Opacity;
         if (igs->SMask) {
-            pdfi_trans_set_mask(ctx, igs, csel);
+            code = pdfi_trans_set_mask(ctx, igs, csel);
         }
     }
 
-    return 0;
+    return code;
 }
